﻿Imports System.Threading



Public Class BKcomm

    Private WithEvents SerialPort As New System.IO.Ports.SerialPort

    ' variables that define the behavior of BK1786 unit and the used protocol
    Private Const MAX_VOLTAGE = 32      ' max output voltage
    Private Const MAX_CURRENT = 3       ' max output current
    Private Const FRAME_LENGTH = 26     ' frame length
    Private Const VOLT_BYTES = 4        ' number of bytes to represent a voltage
    Private Const AMP_BYTES = 2         ' number of bytes to represent a current

    ' class variables
    Private _isConnected As Boolean = False
    Private _address As Integer = 0
    Private _allowSleepFunction As Boolean = False


    ''' <summary>
    ''' isConnected indicates if the BKinstrument is currently connected
    ''' </summary>
    ''' <value>connected? true/false</value>
    ''' <returns>connected? true/false</returns>
    ''' <remarks></remarks>
    Public ReadOnly Property isConnected As Boolean
        Get
            Return _isConnected
        End Get
    End Property




    ''' <summary>
    ''' address contains the hardware address of the BK unit
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property address As Integer
        Get
            Return _address
        End Get
        Set(value As Integer)
            _address = value
        End Set
    End Property

    Public Property allowSleepFunction As Boolean
        Get
            Return _allowSleepFunction
        End Get
        Set(value As Boolean)
            _allowSleepFunction = value
        End Set
    End Property



    ' Enumerations
    Public Enum bkCommand
        READ_PRODUCT_MODEL = &H31
        REMOTE_CONTROL_MODE = &H20
        SET_OUTPUT_STATE = &H21
        SET_MAX_OUTPUT_VOLTAGE = &H22
        SET_OUTPUT_VOLTAGE = &H23
        SET_OUTPUT_CURRENT = &H24
        READ_PRESET_DATA = &H26
    End Enum

    Public Enum remoteControlMode
        FRONT_PANEL_OPERATION = 0
        REMOTE_OPERATION = 1
    End Enum

    Public Enum outputState
        OUTPUT_OFF = 0
        OUTPUT_ON = 1
    End Enum

    Public Enum conversionUnit
        VOLTAGE = 4
        CURRENT = 2
    End Enum




#Region "BK1786 remote instructions"
    ''' <summary>
    ''' setRemoteControlMode() will set the operation control mode (front panel/control mode)
    ''' </summary>
    ''' <param name="operationMode">the desired operation mode</param>
    ''' <returns>A boolean indicating if the function succeeded or not.</returns>
    ''' <remarks></remarks>
    Public Function setRemoteControlMode(ByVal operationMode As remoteControlMode) As Boolean
        Dim frameToSend() As Byte
        frameToSend = makeFrame(bkCommand.REMOTE_CONTROL_MODE, operationMode)

        Dim frameReceived(0) As Byte
        Return sendFrame(frameToSend, frameReceived)
    End Function



    ''' <summary>
    ''' setOutputMode() will set the output mode (active/inactive)
    ''' </summary>
    ''' <param name="mode">the desired output mode</param>
    ''' <returns>A boolean indicating if the function succeeded or not.</returns>
    ''' <remarks></remarks>
    Public Function enableOutput(ByVal mode As Boolean) As Boolean
        Dim frameToSend() As Byte

        frameToSend = makeFrame(bkCommand.SET_OUTPUT_STATE,
                                IIf(mode, outputState.OUTPUT_ON, outputState.OUTPUT_OFF))

        Dim frameReceived(0) As Byte
        Return sendFrame(frameToSend, frameReceived)
    End Function



    ''' <summary>
    ''' setOutputVoltage() will set the output voltage. The function expects a value in the string format
    ''' </summary>
    ''' <param name="milliVolts">the desired voltage (0-32 V)</param>
    ''' <returns>A boolean indicating if the function succeeded or not.</returns>
    ''' <remarks></remarks>
    Public Function setOutputVoltage(ByVal milliVolts As Integer) As Boolean
        Dim voltageByte() As Byte = numberToByte(milliVolts, conversionUnit.VOLTAGE)

        If UBound(voltageByte) = conversionUnit.VOLTAGE - 1 Then
            Dim frameToSend() As Byte
            frameToSend = makeFrame(bkCommand.SET_OUTPUT_VOLTAGE, voltageByte(0), voltageByte(1), voltageByte(2), voltageByte(3))

            Dim frameReceived(0) As Byte
            Return sendFrame(frameToSend, frameReceived)
        Else
            Return False
        End If
    End Function



    ''' <summary>
    ''' setOutputVoltage() will set the output current. The function expects a value in the string format
    ''' </summary>
    ''' <param name="milliAmperes">the desired current (0-3 A)</param>
    ''' <returns>A boolean indicating if the function succeeded or not.</returns>
    ''' <remarks></remarks>
    Public Function setOutputCurrent(ByVal milliAmperes As Integer) As Boolean
        Dim ampereByte() As Byte = numberToByte(milliAmperes, conversionUnit.CURRENT)

        If UBound(ampereByte) = conversionUnit.CURRENT - 1 Then
            Dim frameToSend() As Byte
            frameToSend = makeFrame(bkCommand.SET_OUTPUT_CURRENT, ampereByte(0), ampereByte(1))

            Dim frameReceived(0) As Byte
            Return sendFrame(frameToSend, frameReceived)
        Else
            Return False
        End If
    End Function



    ''' <summary>
    ''' bkCheckConnection() will check if a connection with the BK could be established.
    ''' </summary>
    ''' <returns>A true/false indicating if the connection was established</returns>
    ''' <remarks></remarks>
    Private Function bkCheckConnection() As Boolean

        Dim frameToSend() As Byte
        frameToSend = makeFrame(bkCommand.READ_PRODUCT_MODEL)

        Dim frameReceived(0) As Byte
        Return sendFrame(frameToSend, frameReceived)
    End Function



    ''' <summary>
    ''' IsOutputActive() will return true when the BK's output is active
    ''' </summary>
    ''' <returns>a boolean to indicate if the output is active or not</returns>
    ''' <remarks>The function will also return true in case of a communication problem!</remarks>
    Public Function IsOutputActive() As Boolean
        Dim isOvercurrent As Boolean
        Dim presentOutputMode As outputState
        Dim presentVoltage, setupVoltage, presentCurrent, setupCurrent As Integer

        If readPresentData(isOvercurrent, presentOutputMode, presentVoltage, setupVoltage, presentCurrent, setupCurrent) Then
            Return presentOutputMode = outputState.OUTPUT_ON
        Else
            Return False
        End If
    End Function



    ''' <summary>
    ''' IsOverCurrentOutput() will return true when the BK is in CC mode (current limiting)
    ''' </summary>
    ''' <returns>A boolean that indicates if the BK is in CC mode or not</returns>
    ''' <remarks>The function will also return true in case of a communication problem!</remarks>
    Public Function IsOverCurrentOutput() As Boolean
        Dim isOvercurrent As Boolean
        Dim presentOutputMode As outputState
        Dim presentVoltage, setupVoltage, presentCurrent, setupCurrent As Integer

        If readPresentData(isOvercurrent, presentOutputMode, presentVoltage, setupVoltage, presentCurrent, setupCurrent) Then
            Return isOvercurrent
        Else
            Return False
        End If
    End Function



    ''' <summary>
    ''' getPresentVoltage() will return the actual output voltage
    ''' </summary>
    ''' <returns>The present output voltage (mV)</returns>
    ''' <remarks>The function will also return true in case of a communication problem!</remarks>
    Public Function getPresentVoltage() As Integer
        Dim isOvercurrent As Boolean
        Dim presentOutputMode As outputState
        Dim presentVoltage, setupVoltage, presentCurrent, setupCurrent As Integer

        If readPresentData(isOvercurrent, presentOutputMode, presentVoltage, setupVoltage, presentCurrent, setupCurrent) Then
            Return presentVoltage
        Else
            Return 0
        End If
    End Function



    ''' <summary>
    ''' getSetupVoltage() will return the preselected output voltage
    ''' </summary>
    ''' <returns>The preselected output voltage (mV)</returns>
    ''' <remarks>The function will also return true in case of a communication problem!</remarks>
    Public Function getSetupVoltage() As Integer
        Dim isOvercurrent As Boolean
        Dim presentOutputMode As outputState
        Dim presentVoltage, setupVoltage, presentCurrent, setupCurrent As Integer

        If readPresentData(isOvercurrent, presentOutputMode, presentVoltage, setupVoltage, presentCurrent, setupCurrent) Then
            Return setupVoltage
        Else
            Return 0
        End If
    End Function



    ''' <summary>
    ''' getPresentCurrent() will return the actual current output
    ''' </summary>
    ''' <returns>The present current output (mA)</returns>
    ''' <remarks>The function will also return true in case of a communication problem!</remarks>
    Public Function getPresentCurrent() As Integer
        Dim isOvercurrent As Boolean
        Dim presentOutputMode As outputState
        Dim presentVoltage, setupVoltage, presentCurrent, setupCurrent As Integer

        If readPresentData(isOvercurrent, presentOutputMode, presentVoltage, setupVoltage, presentCurrent, setupCurrent) Then
            Return presentCurrent
        Else
            Return 0
        End If
    End Function



    ''' <summary>
    ''' getSetupCurrent() will return the preselected output current
    ''' </summary>
    ''' <returns>the preselected output current (mA)</returns>
    ''' <remarks></remarks>
    Public Function getSetupCurrent() As Integer
        Dim isOvercurrent As Boolean
        Dim presentOutputMode As outputState
        Dim presentVoltage, setupVoltage, presentCurrent, setupCurrent As Double

        If readPresentData(isOvercurrent, presentOutputMode, presentVoltage, setupVoltage, presentCurrent, setupCurrent) Then
            Return setupCurrent
        Else
            Return 0
        End If
    End Function



    ''' <summary>
    ''' readPresentData() will read all the present data of the device. The results are passed 
    ''' through the byReference variables.
    ''' </summary>
    ''' <param name="isOvercurrent">is the unit currently in CC mode?</param>
    ''' <param name="presentOutputMode">is the output currently active of not?</param>
    ''' <param name="presentVoltage">The present output voltage</param>
    ''' <param name="setupVoltage">The programmed (desired) voltage</param>
    ''' <param name="presentCurrent">The present output current</param>
    ''' <param name="setupCurrent">The programmed (desired) current</param>
    ''' <returns>A boolean to indicate if the function call succeeded</returns>
    ''' <remarks></remarks>
    Public Function readPresentData(ByRef isOvercurrent As Boolean,
                                    ByRef presentOutputMode As outputState,
                                    ByRef presentVoltage As Integer,
                                    ByRef setupVoltage As Integer,
                                    ByRef presentCurrent As Integer,
                                    ByRef setupCurrent As Integer) As Boolean
        Dim frameToSend() As Byte
        frameToSend = makeFrame(bkCommand.READ_PRESET_DATA)

        Dim frameReceived(0) As Byte
        If sendFrame(frameToSend, frameReceived) Then
            ' Step 1: Analyse power supply state
            Dim code As Byte = frameReceived(9)
            isOvercurrent = ((code And 8) = 8)
            presentOutputMode = ((code And 1) = 1)

            ' Step 2: Analyse present output voltage
            Dim presentVoltageData(3) As Byte
            presentVoltageData(0) = frameReceived(5)
            presentVoltageData(1) = frameReceived(6)
            presentVoltageData(2) = frameReceived(7)
            presentVoltageData(3) = frameReceived(8)
            presentVoltage = byteToValue(presentVoltageData)

            ' Step 3: Analyse setup voltage value
            Dim setupVoltageData(3) As Byte
            setupVoltageData(0) = frameReceived(16)
            setupVoltageData(1) = frameReceived(17)
            setupVoltageData(2) = frameReceived(18)
            setupVoltageData(3) = frameReceived(19)
            setupVoltage = byteToValue(setupVoltageData)

            ' Step 4: Analyse present current voltage
            Dim presentCurrentData(1) As Byte
            presentCurrentData(0) = frameReceived(3)
            presentCurrentData(1) = frameReceived(4)
            presentCurrent = byteToValue(presentCurrentData)

            ' Step 5: Analyse setup current value 
            Dim setupCurrentData(1) As Byte
            setupCurrentData(0) = frameReceived(10)
            setupCurrentData(1) = frameReceived(11)
            setupCurrent = byteToValue(setupCurrentData)

            Return True
        Else
            ' invalid Frame
            Return False
        End If
    End Function
#End Region





#Region "Connection Code"
    ''' <summary>
    ''' This sub will disconnect the BK instrument
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub disconnect()
        SerialPort.Close()
        _isConnected = False
    End Sub



    ''' <summary>
    ''' connect() will make a connection with the BK instrument. 
    ''' 
    ''' Attention: This function may need a few seconds to complete!
    ''' </summary>
    ''' <param name="COM">the COM port</param>
    ''' <param name="baud">the baud rate</param>
    ''' <param name="address">the hardware address</param>
    ''' <returns>A boolean to indicate the function failed</returns>
    ''' <remarks></remarks>
    Public Function connect(ByVal COM As Integer, ByVal baud As Integer, ByVal address As Integer) As Boolean

        Me.address = address

        Try
            disconnect()

            ' Set port properties (hardcoding is OK)
            With SerialPort
                .PortName = "COM" & COM
                .Parity = IO.Ports.Parity.None
                .BaudRate = baud
                .DataBits = 8
                .DtrEnable = True
                .StopBits = IO.Ports.StopBits.One
            End With

            ' This code might trigger an Exception!
            _isConnected = True
            SerialPort.Open()

            ' Now we try to read the port, this may take a few seconds
            ' TODO: add a background thread to handle this part of the code.
            Dim checkCounter As Integer = 0
            Dim connected As Boolean = False

            Do
                Thread.Sleep(250)
                connected = bkCheckConnection()
                checkCounter += 1
            Loop Until (checkCounter > 5 Or connected)

            If connected Then
                _isConnected = True
                Return True
            Else
                _isConnected = False
                SerialPort.Close()
                Return False
            End If
        Catch ex As Exception
            Return False
        End Try
    End Function

#End Region





#Region "Data frame handling & Verification"
    ''' <summary>
    ''' makeFrame() generates a data frame with the desired contents. The checksum is appended
    ''' automatically.
    ''' </summary>
    ''' <param name="command">The byte number of the command to execute</param>
    ''' <param name="byte4">Optional data for byte 4. </param>
    ''' <param name="byte5">Optional data for byte 5. </param>
    ''' <param name="byte6">Optional data for byte 6. </param>
    ''' <param name="byte7">Optional data for byte 7. </param>
    ''' <returns>A byte array with 26 positions</returns>
    ''' <remarks></remarks>
    Private Function makeFrame(ByVal command As Byte,
                                Optional ByVal byte4 As Integer = 0,
                                Optional ByVal byte5 As Integer = 0,
                                Optional ByVal byte6 As Integer = 0,
                                Optional ByVal byte7 As Integer = 0) As Byte()

        ' Construct the request frame
        Dim newFrame(FRAME_LENGTH - 1) As Byte
        newFrame(0) = &HAA              ' A frame always starts with AA
        newFrame(1) = address           ' The hardware address (hardcoded to 0)
        newFrame(2) = command           ' The command to execute
        newFrame(3) = byte4             ' optional data byte 4
        newFrame(4) = byte5             ' optional data byte 5
        newFrame(5) = byte6             ' optional data byte 6
        newFrame(6) = byte7             ' optional data byte 7
        newFrame(FRAME_LENGTH - 1) =
            calcCheckSum(newFrame)      ' append checksum data

        Return newFrame
    End Function



    ''' <summary>
    ''' sendRequest() will send a data request to the BK instrument. The
    ''' function returns the response from the instrument.
    ''' </summary>
    ''' <param name="frameToSend">The frame to send</param>
    ''' <returns>A byte-array containing the response</returns>
    ''' <remarks></remarks>
    Private Function sendFrame(ByVal frameToSend() As Byte, ByRef frameReceived() As Byte) As Boolean

        ' here we will send the code to our bk unit
        Try
            If isConnected Then
                If (allowSleepFunction = True) Then Thread.Sleep(100)
                ' Write the frame to the port

                SerialPort.Write(frameToSend, 0, FRAME_LENGTH)

                Debug.Print("TX: " & ByteArrayToString(frameToSend))

                ' Get the number of received bytes and define the buffer
                Dim bufferSize = SerialPort.BytesToRead
                ReDim frameReceived(bufferSize - 1)

                ' Loop through all the received data, add it to the buffer
                For i = 0 To bufferSize - 1
                    frameReceived(i) = SerialPort.ReadByte()
                Next

                Debug.Print("RX: " & ByteArrayToString(frameReceived))
                Return isDataValid(frameReceived)
            Else
                Return False
            End If
        Catch ex As Exception
            Return False
        End Try
    End Function



    ''' <summary>
    ''' isDataValid() will verify the checksum of the given data frame.
    ''' </summary>
    ''' <param name="frame">The complete data frame</param>
    ''' <returns>A boolean that indicates if the frame has the right checksum</returns>
    ''' <remarks></remarks>
    Public Function isDataValid(ByVal frame() As Byte) As Boolean

        Dim valid As Boolean = True

        If UBound(frame) = FRAME_LENGTH - 1 Then
            ' The frame has the expected length of 26 bytes
            Dim checksum As Integer = calcCheckSum(frame)
            If checksum <> frame(FRAME_LENGTH - 1) Then
                ' The checksum doesn't match the data
                Debug.Print("Eval: Error (checksum)")
                valid = False
            Else
                Debug.Print("Eval: OK")
                valid = True
            End If
        Else
            ' The frame doesn't have the expected length
            valid = False
            Debug.Print("Eval: Error (data length)")
        End If

        ' Pass the result
        Return valid
    End Function



    ''' <summary>
    ''' calcCheckSum() calculates the checksum for the given frame.
    ''' </summary>
    ''' <param name="aFrame">The frame you want to calculate the checksum for</param>
    ''' <returns>The calculated checksum (the sum of the first 24 bytes, modulus 256)</returns>
    ''' <remarks></remarks>
    Private Function calcCheckSum(ByVal aFrame() As Byte) As Integer
        Dim totSum As Integer = 0

        For i = 0 To UBound(aFrame) - 2
            totSum += aFrame(i)
        Next

        Dim checksum As Integer = totSum Mod 256
        Return checksum

    End Function
#End Region





#Region "Helper & Conversion Functions"
    ''' <summary>
    ''' ByteArrayToString() will convert an array with bytes to a hexadecimal String. It is used to show
    ''' the data packets during debugging.
    ''' </summary>
    ''' <param name="byteArray">An array containing the hexadecimal bytes.</param>
    ''' <returns>A hexadecimal string</returns>
    ''' <remarks></remarks>
    Private Function ByteArrayToString(ByVal byteArray() As Byte) As String
        Return BitConverter.ToString(byteArray)
    End Function



    ''' <summary>
    ''' byteToValue converts a number of bytes to a decimal number. It is used to convert the
    ''' milliVolt and milliAmperes values to an integer number.
    ''' </summary>
    ''' <param name="bytevalue"></param>
    ''' <returns>The calculated value (milli)</returns>
    ''' <remarks></remarks>
    Private Function byteToValue(ByVal bytevalue() As Byte) As Integer

        Dim hexCode As String = ""

        For i As Integer = UBound(bytevalue) To 0 Step -1
            Dim hexCodeCharacter As String = Conversion.Hex(bytevalue(i))
            hexCode += hexCodeCharacter.PadLeft(2, "0")
        Next i

        Return Integer.Parse(hexCode, Globalization.NumberStyles.HexNumber)
    End Function



    ''' <summary>
    ''' getMaxValue gets the maximal specification for the selected unit.
    ''' </summary>
    ''' <param name="unit">the unit to get the maximal specification for</param>
    ''' <returns>The maximal value</returns>
    ''' <remarks></remarks>
    Private Function getMaxValue(ByVal unit As conversionUnit) As Integer
        If unit = conversionUnit.CURRENT Then
            Return MAX_CURRENT
        Else
            Return MAX_VOLTAGE
        End If
    End Function



    ''' <summary>
    ''' getHexByteLength gets the number of bytes that are used to represent mV or mA.
    ''' </summary>
    ''' <param name="unit">The unit to get the number of bytes for</param>
    ''' <returns>The number of bytes</returns>
    ''' <remarks></remarks>
    Private Function getHexByteLength(ByVal unit As conversionUnit) As Integer
        If unit = conversionUnit.CURRENT Then
            Return AMP_BYTES
        Else
            Return VOLT_BYTES
        End If
    End Function


    ''' <summary>
    ''' getCounterStart() is a helper function.
    ''' </summary>
    ''' <param name="unit"></param>
    ''' <returns></returns>
    ''' <remarks>TODO: this function should be avoidable</remarks>
    Private Function getCounterStart(ByVal unit As conversionUnit) As Integer
        If unit = conversionUnit.VOLTAGE Then
            Return 7
        Else
            Return 3
        End If
    End Function



    Private Function numberToByte(ByVal milliValue As Integer, ByVal desiredUnit As conversionUnit) As Byte()

        Dim conversionData(desiredUnit - 1) As Byte

        ' Check 1: make sure last digit is always 0
        milliValue = 10 * Int(milliValue / 10)

        ' Check 2: convert value to hexadecimal number
        Dim hexValue As String = Conversion.Hex(milliValue)
        hexValue = hexValue.PadLeft(2 * getHexByteLength(desiredUnit), "0")

        ' Check 3: now we process value into a byte array
        Dim byteCounter As Integer = 0
        For i = getCounterStart(desiredUnit) To 1 Step -2
            conversionData(byteCounter) =
                Long.Parse(Mid(hexValue, i, 2), Globalization.NumberStyles.HexNumber)
            byteCounter += 1
        Next

        Return conversionData
    End Function



    ''' <summary>
    ''' V_to_mV converts Volts to milliVolts
    ''' </summary>
    ''' <param name="volts">V value</param>
    ''' <returns>mV value</returns>
    ''' <remarks></remarks>
    Public Shared Function V_to_mV(ByVal volts As Double) As Integer
        Return volts * 1000
    End Function



    ''' <summary>
    ''' A_to_mA converts milliAmpères to Amperes
    ''' </summary>
    ''' <param name="amperes">A value</param>
    ''' <returns>mA value</returns>
    ''' <remarks></remarks>
    Public Shared Function A_to_mA(ByVal amperes As Double) As Integer
        Return amperes * 1000
    End Function



    ''' <summary>
    ''' mV_to_V converts milliVolts to Volts
    ''' </summary>
    ''' <param name="milliVolts">mV value</param>
    ''' <returns>V value</returns>
    ''' <remarks></remarks>
    Public Shared Function mV_to_V(ByVal milliVolts As Integer) As Double
        Return milliVolts / 1000
    End Function



    ''' <summary>
    ''' mA_to_A converts milliAmpères to Ampères
    ''' </summary>
    ''' <param name="milliAmperes">mA value</param>
    ''' <returns>A value</returns>
    ''' <remarks></remarks>
    Public Shared Function mA_to_A(ByVal milliAmperes As Integer) As Double
        Return milliAmperes / 1000
    End Function

#End Region
End Class
